@@ -12,23 +12,30 @@ from unittest import TestCase, mock
 from click.testing import CliRunner
 
 from mbed_tools.cli.build import build
-from mbed_tools.project._internal.project_data import CMAKE_CONFIG_FILE_PATH, CMAKE_BUILD_DIR
+from mbed_tools.project._internal.project_data import BUILD_DIR
+from mbed_tools.build.config import CMAKE_CONFIG_FILE
+
+
+DEFAULT_BUILD_ARGS = ["-t", "GCC_ARM", "-m", "K64F"]
+DEFAULT_BUILD_SUBDIR = pathlib.Path("K64F", "develop", "GCC_ARM")
 
 
 @contextlib.contextmanager
-def mock_project_directory(program, mbed_config_exists=False, build_tree_exists=False):
+def mock_project_directory(
+    program, mbed_config_exists=False, build_tree_exists=False, build_subdir=DEFAULT_BUILD_SUBDIR
+):
     with TemporaryDirectory() as tmp_dir:
         root = pathlib.Path(tmp_dir, "test-project")
         root.mkdir()
         program.root = root
-        program.files.cmake_build_dir = root / CMAKE_BUILD_DIR
-        program.files.cmake_config_file = root / CMAKE_CONFIG_FILE_PATH
+        program.files.cmake_build_dir = root / BUILD_DIR / build_subdir
+        program.files.cmake_config_file = root / BUILD_DIR / build_subdir / CMAKE_CONFIG_FILE
         if mbed_config_exists:
-            program.files.cmake_config_file.parent.mkdir(exist_ok=True)
+            program.files.cmake_config_file.parent.mkdir(exist_ok=True, parents=True)
             program.files.cmake_config_file.touch(exist_ok=True)
 
         if build_tree_exists:
-            program.files.cmake_build_dir.mkdir(exist_ok=True)
+            program.files.cmake_build_dir.mkdir(exist_ok=True, parents=True)
 
         yield
 
@@ -42,9 +49,9 @@ class TestBuildCommand(TestCase):
         self, generate_config, mbed_program, build_project, generate_build_system
     ):
         runner = CliRunner()
-        runner.invoke(build)
+        runner.invoke(build, DEFAULT_BUILD_ARGS)
 
-        mbed_program.from_existing.assert_called_once_with(pathlib.Path(os.getcwd()))
+        mbed_program.from_existing.assert_called_once_with(pathlib.Path(os.getcwd()), DEFAULT_BUILD_SUBDIR)
 
     def test_searches_for_mbed_program_at_user_defined_project_root(
         self, generate_config, mbed_program, build_project, generate_build_system
@@ -52,35 +59,29 @@ class TestBuildCommand(TestCase):
         project_path = "my-blinky"
 
         runner = CliRunner()
-        runner.invoke(build, ["-p", project_path])
+        runner.invoke(build, ["-p", project_path, *DEFAULT_BUILD_ARGS])
 
-        mbed_program.from_existing.assert_called_once_with(pathlib.Path(project_path))
+        mbed_program.from_existing.assert_called_once_with(pathlib.Path(project_path), DEFAULT_BUILD_SUBDIR)
 
     def test_calls_generate_build_system_if_build_tree_nonexistent(
         self, generate_config, mbed_program, build_project, generate_build_system
     ):
         program = mbed_program.from_existing()
         with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=False):
+            generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+
             runner = CliRunner()
-            runner.invoke(build, ["-m", "k64f", "-t", "gcc_arm"])
+            runner.invoke(build, DEFAULT_BUILD_ARGS)
 
             generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")
 
-    def test_generate_build_system_not_called_if_build_tree_exists(
-        self, generate_config, mbed_program, build_project, generate_build_system
-    ):
-        program = mbed_program.from_existing()
-        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
-            runner = CliRunner()
-            runner.invoke(build)
-
-            generate_build_system.assert_not_called()
-
     def test_generate_config_called_if_config_script_nonexistent(
         self, generate_config, mbed_program, build_project, generate_build_system
     ):
         program = mbed_program.from_existing()
-        with mock_project_directory(program, mbed_config_exists=False):
+        with mock_project_directory(
+            program, mbed_config_exists=False, build_subdir=pathlib.Path("K64F", "develop", "GCC_ARM")
+        ):
             target = "k64f"
             toolchain = "gcc_arm"
 
@@ -115,25 +116,36 @@ class TestBuildCommand(TestCase):
             self.assertIsNotNone(result.exception)
             self.assertRegex(result.output, "--mbed-target")
 
-    def test_build_system_regenerated_when_target_and_toolchain_passed(
+    def test_raises_if_target_identifier_not_int(
         self, generate_config, mbed_program, build_project, generate_build_system
     ):
-        program = mbed_program.from_existing()
-        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
-            toolchain = "gcc_arm"
-            target = "k64f"
+        target = "K64F[foo]"
 
-            runner = CliRunner()
-            runner.invoke(build, ["-t", toolchain, "-m", target])
+        result = CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm"])
+        self.assertIsNotNone(result.exception)
+        self.assertRegex(result.output, "ID")
 
-            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
-            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")
+    def test_raises_if_target_identifier_negative(
+        self, generate_config, mbed_program, build_project, generate_build_system
+    ):
+        target = "K64F[-1]"
+
+        result = CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm"])
+        self.assertIsNotNone(result.exception)
+        self.assertRegex(result.output, "ID")
 
     def test_build_system_regenerated_when_mbed_os_path_passed(
         self, generate_config, mbed_program, build_project, generate_build_system
     ):
         program = mbed_program.from_existing()
-        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
+        with mock_project_directory(
+            program,
+            mbed_config_exists=True,
+            build_tree_exists=True,
+            build_subdir=pathlib.Path("K64F", "develop", "GCC_ARM"),
+        ):
+            generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+
             toolchain = "gcc_arm"
             target = "k64f"
             mbed_os_path = "./extern/mbed-os"
@@ -144,11 +156,71 @@ class TestBuildCommand(TestCase):
             generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
             generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")
 
-    def test_build_folder_removed_when_clean_flag_passed(
+    def test_custom_targets_location_used_when_passed(
         self, generate_config, mbed_program, build_project, generate_build_system
     ):
         program = mbed_program.from_existing()
         with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
+            toolchain = "gcc_arm"
+            target = "k64f"
+            custom_targets_json_path = pathlib.Path("custom", "custom_targets.json")
+
+            runner = CliRunner()
+            runner.invoke(build, ["-t", toolchain, "-m", target, "--custom-targets-json", custom_targets_json_path])
+
+            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
+            self.assertEqual(program.files.custom_targets_json, custom_targets_json_path)
+
+    def test_app_config_used_when_passed(
+        self, generate_config, mbed_program, build_project, generate_build_system
+    ):
+        program = mbed_program.from_existing()
+        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
+            toolchain = "gcc_arm"
+            target = "k64f"
+            app_config_path = pathlib.Path("alternative_config.json")
+
+            runner = CliRunner()
+            runner.invoke(build, ["-t", toolchain, "-m", target, "--app-config", app_config_path])
+
+            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
+            self.assertEqual(program.files.app_config_file, app_config_path)
+
+    def test_profile_used_when_passed(
+        self, generate_config, mbed_program, build_project, generate_build_system
+    ):
+        program = mbed_program.from_existing()
+        mbed_program.reset_mock()  # clear call count from previous line
+
+        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
+            generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+
+            toolchain = "gcc_arm"
+            target = "k64f"
+            profile = "release"
+
+            runner = CliRunner()
+            runner.invoke(build, ["-t", toolchain, "-m", target, "--profile", profile])
+
+            mbed_program.from_existing.assert_called_once_with(
+                pathlib.Path(os.getcwd()),
+                pathlib.Path(target.upper(), profile, toolchain.upper())
+            )
+            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
+            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, profile)
+
+    def test_build_folder_removed_when_clean_flag_passed(
+        self, generate_config, mbed_program, build_project, generate_build_system
+    ):
+        program = mbed_program.from_existing()
+        with mock_project_directory(
+            program,
+            mbed_config_exists=True,
+            build_tree_exists=True,
+            build_subdir=pathlib.Path("K64F", "develop", "GCC_ARM"),
+        ):
+            generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+
             toolchain = "gcc_arm"
             target = "k64f"
 
@@ -160,30 +232,77 @@ class TestBuildCommand(TestCase):
             self.assertFalse(program.files.cmake_build_dir.exists())
 
     @mock.patch("mbed_tools.cli.build.flash_binary")
-    @mock.patch("mbed_tools.cli.build.find_connected_device")
-    def test_build_flash_option(
-        self, mock_find_device, flash_binary, generate_config, mbed_program, build_project, generate_build_system
+    @mock.patch("mbed_tools.cli.build.find_all_connected_devices")
+    def test_build_flash_options_bin_target(
+        self,
+        mock_find_devices,
+        flash_binary,
+        generate_config,
+        mbed_program,
+        build_project,
+        generate_build_system,
     ):
+        # A target with bin images does not need OUTPUT_EXT defined
+        generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+        mock_find_devices.return_value = [mock.MagicMock()]
         runner = CliRunner()
-        runner.invoke(build, ["--flash"])
-        flash_binary.assert_called_once()
+        runner.invoke(build, ["--flash", *DEFAULT_BUILD_ARGS])
+        call_args = flash_binary.call_args
+        args, kwargs = call_args
+        flash_binary.assert_called_once_with(args[0], args[1], args[2], args[3], False)
 
     @mock.patch("mbed_tools.cli.build.flash_binary")
-    @mock.patch("mbed_tools.cli.build.find_connected_device")
-    def test_build_flash_and_hex_file_options(
-        self, mock_find_device, flash_binary, generate_config, mbed_program, build_project, generate_build_system
+    @mock.patch("mbed_tools.cli.build.find_all_connected_devices")
+    def test_build_flash_options_hex_target(
+        self,
+        mock_find_devices,
+        flash_binary,
+        generate_config,
+        mbed_program,
+        build_project,
+        generate_build_system,
     ):
+        generate_config.return_value = [{"OUTPUT_EXT": "hex"}, mock.MagicMock()]
+        mock_find_devices.return_value = [mock.MagicMock()]
         runner = CliRunner()
-        runner.invoke(build, ["--flash", "--hex-file"])
+        runner.invoke(build, ["--flash", *DEFAULT_BUILD_ARGS])
         call_args = flash_binary.call_args
         args, kwargs = call_args
         flash_binary.assert_called_once_with(args[0], args[1], args[2], args[3], True)
 
-    def test_build_only_hex_file_option(self, generate_config, mbed_program, build_project, generate_build_system):
+    @mock.patch("mbed_tools.cli.build.flash_binary")
+    @mock.patch("mbed_tools.cli.build.find_all_connected_devices")
+    def test_build_flash_both_two_devices(
+        self,
+        mock_find_devices,
+        flash_binary,
+        generate_config,
+        mbed_program,
+        build_project,
+        generate_build_system,
+    ):
+        generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+        mock_find_devices.return_value = [mock.MagicMock(), mock.MagicMock()]
         runner = CliRunner()
-        result = runner.invoke(build, ["--hex-file"])
+        runner.invoke(build, ["--flash", *DEFAULT_BUILD_ARGS])
+        self.assertEqual(flash_binary.call_count, 2)
 
-        self.assertRegex(result.output, "-f/--flash")
+    @mock.patch("mbed_tools.cli.build.flash_binary")
+    @mock.patch("mbed_tools.cli.build.find_connected_device")
+    def test_build_flash_only_identifier_device(
+        self,
+        mock_find_device,
+        flash_binary,
+        generate_config,
+        mbed_program,
+        build_project,
+        generate_build_system,
+    ):
+        generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
+        mock_find_device.return_value = mock.MagicMock()
+        runner = CliRunner()
+        runner.invoke(build, ["--flash", "-m", "K64F[1]", "-t", "GCC_ARM"])
+        self.assertEqual(flash_binary.call_count, 1)
 
     @mock.patch("mbed_tools.cli.build.terminal")
     @mock.patch("mbed_tools.cli.build.find_connected_device")
@@ -194,10 +313,11 @@ class TestBuildCommand(TestCase):
         serial_port = "tty.k64f"
         baud = 115200
         mock_find_device.return_value = mock.Mock(serial_port=serial_port)
+        generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
 
         CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm", "--sterm", "--baudrate", baud])
 
-        mock_find_device.assert_called_once_with(target)
+        mock_find_device.assert_called_once_with(target, None)
         mock_terminal.run.assert_called_once_with(serial_port, baud)
 
     @mock.patch("mbed_tools.cli.build.terminal")
@@ -208,6 +328,7 @@ class TestBuildCommand(TestCase):
         target = "K64F"
         serial_port = None
         mock_find_device.return_value = mock.Mock(serial_port=serial_port)
+        generate_config.return_value = [mock.MagicMock(), mock.MagicMock()]
 
         output = CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm", "--sterm"])
         self.assertEqual(type(output.exception), SystemExit)
